import json
from typing import Optional, Dict, Any
from ..settings import Settings
from ..serializable import Serializable
from .. import utils, ai

FileCache: "FileCacheObj"

def init():
    global FileCache
    FileCache = FileCacheObj.load()
    ai.set_file_cache(FileCache)

class FileCacheObj(Serializable):
    """Maps file hashes to OpenAI file_ids, and support our dynamic data serialization."""
    
    def __init__(self, cache: Dict[str, str] = {}):
        self._cache = cache
        self.settings = Settings() # type: ignore
        self.path = utils.get_summawise_dir() / f"file_cache.{self.settings.data_mode.ext()}"

    def set_hash_file_id(self, hash: str, file_id: str):
        self._cache[hash] = file_id

    def get_file_id_by_hash(self, hash: str) -> Optional[str]:
        return self._cache.get(hash)

    @classmethod
    def load(cls) -> "FileCacheObj":
        # create empty FileCache
        cache = cls()
        settings = cache.settings

        # if the file doesn't exist, save empty version by default
        if not cache.path.exists():
            cache.save()
            return cache

        # if the file does exist, use functionality fom Serializable base class to establish populated object
        return cache.from_file(cache.path, settings.data_mode)

    def save(self):
        data_mode = self.settings.data_mode
        compress = self.settings.compression
        self.save_to_file(self.path, data_mode, compress, pretty_json = True)

    @classmethod
    def from_json(cls, json_str: str) -> "FileCacheObj":
        cache_dict = json.loads(json_str)
        cache_dict = FileCacheObj.filter_dict(cache_dict)
        cache = FileCacheObj(cache_dict)
        return cache

    def to_json(self, pretty: bool = False) -> str:
        return json.dumps(self._cache, indent = 4 if pretty else None)

    @staticmethod
    def filter_dict(obj: Dict[str, Any]) -> Dict[str, str]:
        """Returns a dict in which any values that aren't strings are removed."""
        return {
            k: v for k, v in obj.items() \
            if isinstance(v, str)
        }



from pathlib import Path
from .metadata import FileMetadata
from . import utils as FileUtils
from .. import ai, utils
from ..settings import Settings

def process_dir(dir_path: Path, delete: bool = True) -> str:
    _ = delete

    files = FileUtils.list_files(dir_path)
    filtered = FileUtils.filter_files(files) # TODO(justin): look into improving/validating approach
    try:
        print(f"Directory scan located and validated {filtered.valid_count}/{filtered.total_count} files.")
        vector_store = ai.create_vector_store(dir_path.name, filtered.files)
        vector_store_id = vector_store.id
        print(f"Vector store created with ID: {vector_store_id}")
    except Exception as ex:
        raise Exception(f"Error creating vector store [{type(ex)}]: {ex}")

    return vector_store_id

def process_file(file_path: Path, delete: bool = False) -> str:
    # TODO(justin): 
    # - add archive support (.zip, .tar.gz) - extract, call process_dir
    # - maybe add automatic extraction of text from pdf or html (undecided)

    # use settings class (singleton)
    settings = Settings() # type: ignore

    # file extension to cache file metadata (json or bin)
    ext = settings.data_mode.ext()

    metadata = FileMetadata.create_from_path(file_path)
    hash = metadata.hash
    hash_path = utils.fp(utils.get_summawise_dir() / "files" / f"{hash}.{ext}")

    vector_store_id: str = ""
    if not hash_path.exists():
        try:
            vector_store = ai.create_vector_store(file_path.stem, [file_path])
            metadata.vector_store_id = vector_store.id
            metadata.save_to_file(
                file_path = hash_path,
                mode = settings.data_mode,
                compress = settings.compression,
                pretty_json = True
            )
            vector_store_id = metadata.vector_store_id
            print(f"Vector store created with ID: {vector_store_id}")
        except Exception as ex:
            raise Exception(f"Error creating vector store [{type(ex)}]: {ex}")
    else:
        metadata = FileMetadata.from_file(hash_path, settings.data_mode)
        vector_store_id = metadata.vector_store_id
        print(f"Restored vector store ID from cache: {vector_store_id}")

    if delete and file_path.exists():
        file_path.unlink()

    return vector_store_id



import click
from typing import Optional
from .. import ai, utils
from ..assistants import Assistant
from ..settings import Settings

@click.group()
def assistant():
    """Commands related to managing assistants."""
    pass

@assistant.command()
def list():
    """List your available assistants."""
    settings = Settings() # type: ignore
    for idx, assistant in enumerate(settings.assistants, start = 1):
        print(f"{idx}) {assistant.name}")

@assistant.command()
@click.option("-n", "--name", prompt = True, help = "The name of the assistant.")
@click.option("-i", "--instructions", prompt = True, help = "Instructions for the assistant.")
@click.option("-m", "--model", default = None, help = "The model to use.")
@click.option("-d", "--description", default = None, help = "Description of the assistant.")
@click.option("-fs", "--file_search", is_flag = True, help = "Enable file search capability.")
@click.option("-ic", "--interpret_code", is_flag = True, help = "Enable code interpretation capability.")
@click.option("-rwj", "--respond_with_json", is_flag = True, help = "Responses should be in valid JSON format.")
@click.option("--temperature", default = None, type = float, help = "Temperature setting for the model.")
@click.option("--top_p", default = None, type = float, help = "Top-p setting for the model.")
def create(
    name: str,
    instructions: str,
    file_search: bool,
    interpret_code: bool,
    respond_with_json: bool,
    model: Optional[str],
    description: Optional[str],
    temperature: Optional[float],
    top_p: Optional[float]
) -> None:
    """Create a new assistant."""
    settings = Settings() # type: ignore

    assistant = settings.assistants.get_by_name(name)
    if assistant:
        print("An assistant with that name already exists.")
        return

    model = model or settings.model
    assistant = Assistant(
        model = model,
        name = name,
        instructions = instructions,
        file_search = file_search,
        interpret_code = interpret_code,
        respond_with_json = respond_with_json,
        description = description,
        temperature = temperature,
        top_p = top_p
    )

    api_assistant = ai.create_assistant(**assistant.to_create_params())
    assistant.apply_api_obj(api_assistant)
    settings.assistants.append(assistant)
    settings.save()

    print(f"Your new assistant has been created successfully.\nName: {name}\nID: {assistant.id}")
    return

@assistant.command()
@click.argument("assistant_id", type = str)
def delete(assistant_id: str):
    """
    Delete an assistant.\n
    Provided ID can be the assistant name, id, or number from 'list'.
    """
    settings = Settings() # type: ignore
    idx = -1

    if assistant_id.startswith("asst_"):
        _, idx = settings.assistants.get_by_id(assistant_id)
        if idx == -1:
            print("Failed to identify assistant to delete.")
            return

    if idx == -1:
        assistant = settings.assistants.get_by_name(assistant_id)
        if assistant:
            _, idx = settings.assistants.get_by_id(assistant.id)

    if idx == -1:
        idx = utils.try_parse_int(assistant_id)
        idx = -1 if idx is None else idx - 1

    if idx < 0 or idx > len(settings.assistants) - 1:
        print("Failed to identify assistant to delete.")
        return

    assistant = settings.assistants[idx]
    del settings.assistants[idx]
    settings.save()
    print(f"Deleted assistant successfully: {assistant.name}")



import warnings
from typing import List, Optional, Iterable, Callable, TypeVar, Tuple
from datetime import datetime, timezone
from dataclasses import dataclass, asdict, field
from . import utils
from .errors import MultipleAssistantsFoundError, MissingSortKeyError
from openai.types.beta import Assistant as APIAssistant

T = TypeVar('T')

DEFAULT_MODEL = "gpt-3.5-turbo"

@dataclass
class Assistant:
    id: str = ""
    name: str = ""
    instructions: str = ""
    model: str = DEFAULT_MODEL
    file_search: bool = False
    interpret_code: bool = False
    respond_with_json: bool = False
    description: Optional[str] = None
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    created_at: datetime = field(default_factory = utils.utc_now)

    def __post_init__(self):
        missing_field_warning = lambda x: warnings.warn(f"'Assistant' object is missing expected field {x}.\nThis field should always be provided.", UserWarning)
        if not self.name: missing_field_warning("name")
        if not self.instructions: missing_field_warning("instructions")

    def to_dict(self) -> dict:
        obj = asdict(self)
        return utils.convert_datetimes(obj, converter = utils.converter_ts_int)
    
    def to_create_params(self) -> dict:
        return utils.asdict_exclude(self, {"id", "created_at"})

    def apply_api_obj(self, obj: APIAssistant):
        """Use official library API object to apply/overwrite specific fields to this dataclass."""
        self.id = obj.id
        self.name = obj.name or ""
        self.instructions = obj.instructions or ""
        self.model = obj.model
        self.description = obj.description
        self.top_p = obj.top_p
        self.temperature = obj.temperature
        self.created_at = datetime.fromtimestamp(obj.created_at, timezone.utc)

class AssistantList(List[Assistant]):

    def __init__(
        self, 
        assistants: Iterable[Assistant] = [], 
        sort: bool = False,
        key: Optional[Callable[[Assistant], T]] = None,
        reverse: bool = False
    ):
        if not sort:
            # sort by "created_at" by default (consistency allows use of index as identifier)
            assistants = sorted(assistants, key = lambda a: a.created_at) # type: ignore
            super().__init__(assistants)
            return

        if not key or not callable(key):
            raise MissingSortKeyError("Can't initialize 'AssistantList' in sorted order without callable key.")

        sorted_assistants = sorted(assistants, key = key, reverse = reverse) # type: ignore
        super().__init__(sorted_assistants)

    def to_dict_list(self) -> List[dict]:
        return [a.to_dict() for a in self]

    @staticmethod
    def from_dict_list(objects: List[dict], **kwargs) -> "AssistantList":
        assistants = [Assistant(**obj) for obj in objects]
        return AssistantList(assistants, **kwargs)

    def list_by_name(self, name: str, case_sensitive: bool = False) -> List[Assistant]:
        t = lambda s: s if case_sensitive else s.lower() # transform func
        return [assistant for assistant in self if t(assistant.name) == t(name)]

    def get_by_name(self, name: str, default: Optional[Assistant] = None, case_sensitive: bool = False) -> Optional[Assistant]:
        assistants = self.list_by_name(name, case_sensitive = case_sensitive)
        if len(assistants) == 0:
            return default
        if len(assistants) > 1:
            raise MultipleAssistantsFoundError("name", name)
        return assistants[0]

    def get_by_id(self, id: str) -> Tuple[Optional[Assistant], int]:
        for idx, assistant in enumerate(self):
            if assistant.id == id:
                return assistant, idx
        return None, -1

TranscriptAnalyzer: Assistant = Assistant(
    name = "Transcript Analysis Assistant",
    instructions = "You are an assistant that summarizes video transcripts and answers questions about them.",
    file_search = True
)

DEFAULT_ASSISTANTS: List[Assistant] = [TranscriptAnalyzer]



import json
from typing import TypeVar, Type
from pathlib import Path
from dataclasses import is_dataclass, asdict
from .data import DataMode
from .files import utils as FileUtils

ST = TypeVar("ST", bound = "Serializable")

class Serializable:

    def save_to_file(
        self, 
        file_path: Path, 
        mode: DataMode = DataMode.JSON, 
        compress: bool = False,
        pretty_json: bool = False
    ):
        if mode == DataMode.JSON:
            json_str = self.to_json(pretty_json)
            FileUtils.write_str(file_path, json_str, compress)
        elif mode == DataMode.BIN:
            FileUtils.save_object(file_path, self, compress)

    @classmethod
    def from_file(
        cls: Type[ST], 
        file_path: Path, 
        mode: DataMode = DataMode.JSON
    ) -> ST:
        if mode == DataMode.JSON:
            json_str = FileUtils.read_str(file_path)
            return cls.from_json(json_str)
        elif mode == DataMode.BIN:
            return FileUtils.load_object(file_path, cls)

    @classmethod
    def from_json(cls: Type[ST], json_str: str) -> ST:
        try:
            data = json.loads(json_str)
            return cls(**data)
        except Exception as ex:
            if not is_dataclass(cls):
                raise NotImplementedError(
                    f"Class '{cls.__name__}' is not a dataclass, and alternative method failed, "
                    f"so it must provide its own implementation of 'from_json'.\nException: {ex}"
                )
            else:
                raise Exception(
                    f"Unexpected exception occurred while invoking '{cls.__name__}.from_json' on @dataclass.\n"
                    f"Exception: {ex}"
                )

    def to_json(self, pretty: bool = False) -> str:
        try:
            obj = asdict(self) if is_dataclass(self) else self.__dict__
            return json.dumps(obj, indent = 4 if pretty else None)
        except Exception as ex:
            if not is_dataclass(self):
                raise NotImplementedError(
                    f"Class '{type(self).__name__}' is not a dataclass, and alternative method failed, "
                    f"so it must provide its own implementation of 'to_json'.\nException: {ex}"
                )
            else:
                raise Exception(
                    f"Unexpected exception occurred while invoking '{type(self).__name__}.to_json' on @dataclass.\n"
                    f"Exception: {ex}"
                )



import click, sys
from .commands import scan, assistant
from .settings import Settings

CONTEXT_SETTINGS = {"max_content_width": sys.maxsize}

@click.group(context_settings = CONTEXT_SETTINGS)
def cli():
    """A tool to use AI to analyze and interact with vectorized data from custom sources of information."""
    pass

def register_commands():
    cli.add_command(scan)
    cli.add_command(assistant)

def main():
    Settings.init()
    register_commands()
    cli()



from dataclasses import dataclass
from pathlib import Path
from . import utils as FileUtils
from ..serializable import Serializable

@dataclass
class FileMetadata(Serializable):
    hash: str
    name: str
    full_path: str
    sz_bytes: int
    created_at: float
    last_modified_at: float
    last_accessed_at: float
    vector_store_id: str = ""

    @classmethod
    def create_from_path(cls, file_path: Path) -> "FileMetadata":
        hash = FileUtils.calculate_hash(file_path)
        stats = file_path.stat()
        sz_bytes = stats.st_size
        created_at = stats.st_ctime
        last_modified_at = stats.st_mtime
        last_accessed_at = stats.st_atime

        return cls(hash, file_path.name, str(file_path), sz_bytes,
                   created_at, last_modified_at, last_accessed_at)    



from typing_extensions import override
from typing import List, Optional, NamedTuple
from pathlib import Path
from openai import OpenAI, AssistantEventHandler
from openai.types.file_object import FileObject
from openai.types.beta import Thread, Assistant, VectorStore 
from openai.types.beta.threads import TextContentBlock, TextDelta, Message, Text
from openai.types.beta.threads.runs import ToolCall
from .files import utils as FileUtils
from .files.cache import FileCacheObj
from . import utils

Client: OpenAI
FileCache: FileCacheObj

class EventHandler(AssistantEventHandler):

    def __init__(self, auto_print: bool = False):
        super().__init__()
        self.auto_print = auto_print
        self.response_text = ""

    @override
    def on_text_created(self, text: Text) -> None:
        _ = text
        if self.auto_print: print("\nsummawise > ", end = "", flush = True)

    @override
    def on_text_delta(self, delta: TextDelta, snapshot: Text) -> None:
        _ = snapshot
        if self.auto_print: print(delta.value, end = "", flush = True)

    @override
    def on_tool_call_created(self, tool_call: ToolCall):
        _ = tool_call

    @override
    def on_message_done(self, message: Message) -> None:
        print() # new line
        for content in message.content:
            # other possibilities: ImageFileContentBlock, ImageURLContentBlock
            if isinstance(content, TextContentBlock):
                self.response_text += content.text.value

def init(api_key: str, verify: bool = True):
    """
    Initializes the OpenAI client with the provided API key.

    Parameters:
        api_key (str): The API key to initialize the OpenAI API client with.
        verify (bool): Verify the API key by making a test request to the API. Default is True.

    Raises:
        ValueError: If the API key is not provided or is not a string.
        openai.AuthenticationError: If the API key is invalid or the verification request fails.
    """
    if not api_key or not isinstance(api_key, str):
        raise ValueError("API key must be a non-empty string.")

    global Client
    if "Client" in globals():
        if Client.api_key == api_key:
            return
    
    Client = OpenAI(api_key = api_key)
    if verify:
        Client.models.list()

class FileInfo(NamedTuple):
    hash: str
    file_id: str
    cached: bool = False

def create_file(file_path: Path) -> FileObject:
    with open(file_path, 'rb') as file:
        file_response = Client.files.create(file = file, purpose = "assistants")
        return file_response

def get_file_infos(files: List[Path]) -> List[FileInfo]:
    file_infos: List[FileInfo] = []
    for file_path in files:

        hash = FileUtils.calculate_hash(file_path)
        assert isinstance(hash, str), \
            "Calculated hash should be of type 'str'. Ensure the 'intdigest' parameter is set to false."

        file_id = FileCache.get_file_id_by_hash(hash)
        if file_id is None:
            # not cached, upload new file
            file = create_file(file_path)
            info = FileInfo(hash, file.id,)
            file_infos.append(info)
            FileCache.set_hash_file_id(hash, file.id)
        else:
            # use cached file id
            info = FileInfo(hash, file_id, True)
            file_infos.append(info)

    FileCache.save()
    return file_infos

def create_vector_store(name: str, file_paths: List[Path]) -> VectorStore:
    print(f"Creating vector store with {len(file_paths)} file(s).", end = " ")
    file_infos = get_file_infos(file_paths)
    file_ids = [info.file_id for info in file_infos]

    cached_count = sum(1 for info in file_infos if info.cached)
    print(f"[{cached_count} file(s) already cached]" if cached_count > 0 else "")

    vector_store = Client.beta.vector_stores.create(name = name, file_ids = file_ids)
    return vector_store

def create_assistant(
    model: str,
    name: str, 
    instructions: str, 
    file_search: bool = False,
    interpret_code: bool = False,
    respond_with_json: bool = False,
    description: Optional[str] = None,
    temperature: Optional[float] = None,
    top_p: Optional[float] = None
) -> Assistant:
    tools = []
    if file_search:
        tools.append({"type": "file_search"})
    if interpret_code:
        tools.append({"type": "code_interpreter"})
    response_format = { "type": "json_object" } if respond_with_json else "auto"
    assistant = Client.beta.assistants.create(
        model = model,
        name = name,
        instructions = instructions,
        tools = tools,
        description = description,
        metadata = { "created_by": utils.package_name() },
        top_p = top_p,
        temperature = temperature,
        response_format = response_format, # type: ignore
    )
    return assistant

def get_assistant(id: str) -> Assistant:
    return Client.beta.assistants.retrieve(id)

def create_thread(vector_store_ids: List[str]) -> Thread:
    thread = Client.beta.threads.create(
        tool_resources = {"file_search": {"vector_store_ids": vector_store_ids}}
    )
    return thread

def get_thread_response(thread_id: str, assistant_id: str, prompt: str, auto_print: bool = False) -> str:
    Client.beta.threads.messages.create(thread_id = thread_id, content = prompt, role = "user")

    event_handler = EventHandler(auto_print = auto_print)
    with Client.beta.threads.runs.stream(
        thread_id = thread_id,
        assistant_id = assistant_id,
        event_handler = event_handler
    ) as stream:
        stream.until_done()

    return event_handler.response_text

def set_file_cache(file_cache: FileCacheObj):
    global FileCache
    FileCache = file_cache



import json, os, copy, warnings
from dataclasses import dataclass, fields, field
from typing import Set, Optional, Dict, List, Any, ClassVar, Tuple
from openai import AuthenticationError, BadRequestError
from pathlib import Path
from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter
from . import utils, ai
from .utils import Singleton
from .data import DataMode
from .files import utils as FileUtils
from .assistants import DEFAULT_ASSISTANTS, DEFAULT_MODEL, Assistant, AssistantList
    
@dataclass
class Settings(metaclass = Singleton):
    api_key: str
    assistant_id: str = field(repr = False) # NOTE(justin): deprecated in version 0.3.0
    assistants: AssistantList
    model: str
    compression: bool
    data_mode: DataMode

    DEFAULT_MODEL: ClassVar[str] = DEFAULT_MODEL
    DEFAULT_COMPRESSION: ClassVar[bool] = True
    DEFAULT_DATA_MODE: ClassVar[DataMode] = DataMode.BIN
    DEPRECATED_FIELDS: ClassVar[Set[str]] = {"assistant_id"}

    # NOTE(justin): This class functions as a singleton. Example usage anywhere:
    # settings = Settings() # type: ignore (dismiss warnings related to required arguments)
    # print(settings.to_dict()) # contains info established in main()

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> Tuple["Settings", bool]:
        """
        Create a Settings object from a dictionary representation.

        Parameters:
            data (Dict[str, Any]): A dictionary containing the settings data.

        Returns:
            Tuple["Settings", bool]: A tuple containing the Settings object created from the data and a boolean indicating whether the settings should be updated.
        """
        key_count = len(data)
        expected_key_count = len(fields(Settings)) - len(Settings.DEPRECATED_FIELDS)
        save = key_count != expected_key_count

        assistant_id = data.pop("assistant_id", None) # NOTE(justin): deprecated in version 0.3.0
        assistants = data.pop("assistants", [])
        assistants = AssistantList.from_dict_list(assistants)

        settings = Settings(
            assistant_id = assistant_id,
            assistants = assistants,
            model = data.pop("model", Settings.DEFAULT_MODEL),
            compression = data.pop("compression", Settings.DEFAULT_COMPRESSION),
            data_mode = DataMode(data.pop("data_mode", Settings.DEFAULT_DATA_MODE.value)),
            **data
        )

        # version 0.3.0 backwards compatability (adding multi-assistant support)
        if settings.assistant_id:
            ai.init(settings.api_key, verify = False)
            api_assistant = ai.get_assistant(assistant_id)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                assistant = Assistant()
            assistant.apply_api_obj(api_assistant)
            assert assistant.name and assistant.instructions
            assistants.append(assistant)
            save = True

        return settings, save

    def to_dict(self) -> Dict[str, Any]:
        data = utils.asdict_exclude(self, Settings.DEPRECATED_FIELDS)
        data["assistants"] = self.assistants.to_dict_list()
        data["data_mode"] = self.data_mode.value
        return data

    @staticmethod
    def init() -> "Settings":
        file = Settings.file()
        if file.exists():
            s_str = FileUtils.read_str(file)
            s_dict = json.loads(s_str)
            settings, save = Settings.from_dict(s_dict)
        else:
            settings = Settings.prompt()
            save = True
        
        # initialize openai api
        # TODO(justin): key verification after settings init in main
        ai.init(settings.api_key, verify = False)

        # automatically create default assistants added in future versions
        for da in DEFAULT_ASSISTANTS:
            if not settings.assistants.get_by_name(da.name):
                assistant = copy.copy(da)
                api_assistant = ai.create_assistant(**da.to_create_params())
                assistant.apply_api_obj(api_assistant)
                settings.assistants.append(assistant)
                save = True
        
        if save:
            Settings.save()

        return settings

    @staticmethod
    def save() -> "Settings":
        settings = Settings() # type: ignore
        s_str = json.dumps(settings.to_dict(), indent = 4)
        FileUtils.write_str(Settings.file(), s_str)
        return settings
    
    @staticmethod
    def file() -> Path:
        return utils.get_summawise_dir() / "settings.json"

    @staticmethod
    def prompt() -> "Settings":
        api_key = os.getenv("OPENAI_API_KEY", "")
        api_key = validate_api_key(api_key)
        
        while api_key is None:
            api_key = prompt("Enter your OpenAI API key: ")
            api_key = validate_api_key(api_key)
            if not api_key:
                print("The API key you entered is invalid. Try again.")

        model_completer = WordCompleter(["gpt-4o", "gpt-4", "gpt-4-turbo", "gpt-3.5-turbo"])
        while True:
            try:
                # TODO(justin): add validator to this prompt
                model = prompt(f"Enter the OpenAI model to use [Default: {Settings.DEFAULT_MODEL}]: ", completer = model_completer)
                break
            except BadRequestError as ex:
                if ex.code == "model_not_found":
                    print("The model identifier you entered is invalid. Try again.")
                    continue
                raise ex

        assistants: List[Assistant] = []
        for da in DEFAULT_ASSISTANTS:
            assistant = copy.copy(da)
            api_assistant = ai.create_assistant(**da.to_create_params())
            assistant.apply_api_obj(api_assistant)
            assistants.append(assistant)

        # TODO(justin): maybe add ability to select data mode. possibly a cli option to re-configure settings too.
        # it's not too important, for now it'll default to binary and can be changed manually.

        return Settings(
            api_key = api_key, 
            model = model, 
            assistant_id = "", # NOTE(justin): deprecated in version 0.3.0
            assistants = AssistantList(assistants),
            data_mode = Settings.DEFAULT_DATA_MODE,
            compression = Settings.DEFAULT_COMPRESSION
        )

def validate_api_key(api_key: str) -> Optional[str]:
    try:
        ai.init(api_key)
        return api_key
    except AuthenticationError:
        return None



from enum import Enum
from typing import Optional

class Encoding(Enum):
    """
    Encodings supported by the chardet library, turned into an enum.
    Source: https://github.com/chardet/chardet/blob/9630f2382faa50b81be2f96fd3dfab5f6739a0ef/docs/supported-encodings.rst
    """
    BIG5 = "Big5"
    GB2312 = "GB2312"
    GB18030 = "GB18030"
    EUC_TW = "EUC-TW"
    HZ_GB_2312 = "HZ-GB-2312"
    ISO_2022_CN = "ISO-2022-CN"
    EUC_JP = "EUC-JP"
    SHIFT_JIS = "SHIFT_JIS"
    ISO_2022_JP = "ISO-2022-JP"
    EUC_KR = "EUC-KR"
    ISO_2022_KR = "ISO-2022-KR"
    KOI8_R = "KOI8-R"
    MAC_CYRILLIC = "MacCyrillic"
    IBM855 = "IBM855"
    IBM866 = "IBM866"
    ISO_8859_5_RUS = "ISO-8859-5"
    WINDOWS_1251_RUS = "windows-1251"
    ISO_8859_2_HUN = "ISO-8859-2"
    WINDOWS_1250_HUN = "windows-1250"
    ISO_8859_5_BUL = "ISO-8859-5"
    WINDOWS_1251_BUL = "windows-1251"
    ISO_8859_1_WEU = "ISO-8859-1"
    WINDOWS_1252_WEU = "windows-1252"
    ISO_8859_7_GREEK = "ISO-8859-7"
    WINDOWS_1253_GREEK = "windows-1253"
    ISO_8859_8_HEB_VIS = "ISO-8859-8"
    WINDOWS_1255_HEB_VIS = "windows-1255"
    ISO_8859_8_HEB_LOG = "ISO-8859-8"
    WINDOWS_1255_HEB_LOG = "windows-1255"
    TIS_620 = "TIS-620"
    UTF_32_BE = "UTF-32 BE"
    UTF_32_LE = "UTF-32 LE"
    UTF_32_3412 = "UTF-32 3412"
    UTF_32_2143 = "UTF-32 2143"
    UTF_16_BE = "UTF-16 BE"
    UTF_16_LE = "UTF-16 LE"
    UTF_8 = "UTF-8"
    ASCII = "ASCII"
    
    @classmethod
    def from_string(cls, encoding_str) -> Optional["Encoding"]:
        """
        Get 'Encoding' Enum from chardet string. 
        Returns 'None' if no match is found.
        """
        for encoding in cls:
            if encoding.value.lower() == encoding_str.lower():
                return encoding
        return None



import validators, click, requests, time, sys
from openai.types.beta import VectorStore
from typing import Tuple, Dict, Optional
from prompt_toolkit import prompt
from pathlib import Path
from .. import ai
from ..settings import Settings
from ..assistants import Assistant
from ..web import process_url
from ..files.processing import process_file, process_dir
from ..files import cache as FileCache
from ..errors import NotSupportedError
from ..utils import NumericChoiceValidator, delete_lines
from ..data import DataUnit

@click.command()
@click.argument("user_input", nargs = -1)
def scan(user_input: Tuple[str, ...]):
    """Scan and process the given input (URL or file path), and offer an interactive prompt to inquire about the vectorized data."""
    settings = Settings() # type: ignore
    FileCache.init()

    while True:
        # prompt user for data source if not provided as an argument
        if user_input:
            input_str = "".join(user_input)
        else:
            input_str = prompt("Enter a URL or local file path: ").strip('\'"')

        # invoke process_input func to handle processing of data and retrieve VectorStore ID
        try:
            vector_store_id = process_input(input_str)
            break
        except NotSupportedError as ex:
            print(ex)
            user_input = ()
            continue
        except (requests.RequestException, requests.HTTPError) as ex:
            print(f"An error occurred while sending a web request to the provided URL:\n{ex}")
            user_input = ()
            continue
        except Exception as ex:
            print(f"An unhandled error occurred while processing input:\n{ex}")
            user_input = ()
            continue

    # make sure the VectorStore ID we got seems correct
    try:
        assert len(vector_store_id) > 0, "empty"
        assert vector_store_id.startswith("vs_"), "invalid format"
    except AssertionError as ex:
        print(f"An unknown occurred while processing input: invalid VectorStore ID. ({ex})")
        return

    if len(settings.assistants) > 1:
        # list assistant choices, map numeric index to name
        assistant_choices: Dict[int, Assistant] = {}
        for idx, assistant in enumerate(settings.assistants, start = 1):
            assistant_choices[idx] = assistant
            print(f"{idx}) {assistant.name}")

        # prompt user to select assistant 
        assistant_id = ""
        assistant: Optional[Assistant] = None
        while not (assistant_id and assistant):
            choice = int(prompt("Select an assistant: ", validator = NumericChoiceValidator(assistant_choices.keys())))
            # assistant_id = settings.assistants[assistant_name]
            assistant = assistant_choices[choice]
            assistant_id = assistant.id

        # remove assistant selection menu, output valid choice
        delete_lines(len(assistant_choices) + 2)
        print(f"Using selected assistant: {assistant.name}")
    else:
        assistant_id = settings.assistants[0].id

    # verify vector store validity w/ openai, output some generic info
    processing: bool = True
    vector_store: Optional[VectorStore] = None
    while processing:
        try:
            first_run = vector_store is None
            vector_store = ai.Client.beta.vector_stores.retrieve(vector_store_id) # model dump example: https://pastebin.com/k4fwANdi
            assert vector_store.status != "expired", "VectorStore is expired."

            id, name = vector_store_id, vector_store.name
            total_files = vector_store.file_counts.total
            completed_files = vector_store.file_counts.completed
            pending_files = vector_store.file_counts.in_progress
            processing = vector_store.status != "completed"

            size_str = DataUnit.bytes_to_str(vector_store.usage_bytes)
            info = f"{id} ({name}), {completed_files}/{total_files} file(s), {size_str}"
            if first_run: 
                print(f"Successfully established VectorStore!", end = " ")
                if processing:
                    print(f"Processing data from {pending_files} file(s)...")
                else:
                    print(f"{completed_files}/{total_files} file(s) have already been processed.")

            if processing: 
                time.sleep(2)
                continue

            print(f"VectorStore ready for use. [{info}]")
        except Exception as ex:
            print(f"Failed to validate VectorStore from provided ID (error: {type(ex)}, id: {vector_store_id}):\n{ex}")
            return

    # create thread for this conversation
    try:
        # TODO(justin): change message used to initialize thread based on type of input data/source
        thread = ai.create_thread([vector_store_id])
        print(f"Thread created with ID: {thread.id}")
        print("Generating summary...")
        ai.get_thread_response(thread.id, assistant_id, "Please summarize the transcript.", auto_print = True)
    except Exception as e:
        print(f"Error generating summary: {e}")
        return
    
    print("\nYou can now ask questions about the transcript. Type 'exit' to quit.")
    while True:
        input_str = prompt("\nyou > ")
        conditional_exit(input_str)
        try:
            ai.get_thread_response(thread.id, assistant_id, input_str, auto_print = True)
        except Exception as e:
            print(f"Error in chat: {e}")

def process_input(user_input: str) -> str:
    """Takes user input, attempts to return OpenAI VectorStore ID after processing data."""
    conditional_exit(user_input)

    path = Path(user_input)
    if path.exists():
        if path.is_file():
            return process_file(path)
        elif path.is_dir():
            return process_dir(path)

    if validators.url(user_input):
        return process_url(user_input)

    raise NotSupportedError()

def conditional_exit(user_input: str) -> None:
    """Conditionally exit the program based on the users input."""
    if user_input.lower() in ["exit", "quit", ":q"]:
        if user_input == ":q": print("They should call you Vim Diesel.") # NOTE(justin): this is here to stay
        sys.exit()



import inspect
from enum import Enum
from typing import List

class DataMode(Enum):
    JSON = "json"
    BIN = "binary"

    def ext(self) -> str:
        extensions = {
            DataMode.JSON: "json",
            DataMode.BIN: "bin"
        }
        try:
            return extensions[self]
        except KeyError:
            raise ValueError(f"Unsupported DataMode: {self}")

class DataUnit:
    __excludes__ = ["units"]

    B: int = 1
    KB: int = 2 ** 10
    MB: int = 2 ** 20
    GB: int = 2 ** 30
    TB: int = 2 ** 40

    units: List[str] = []

    @staticmethod
    def _get_units() -> List[str]:
        # "B", "KB", "MB", "GB", "TB"
        if len(DataUnit.units):
            return DataUnit.units
        units = [
            (name, value) for name, value in inspect.getmembers(DataUnit) 
            if not name.startswith("__") 
            and name not in DataUnit.__excludes__ 
            and not callable(value)
        ]
        units = sorted(units, key=lambda x: x[1])
        DataUnit.units = [name for name, _ in units]
        return DataUnit.units

    @staticmethod
    def bytes_to_str(sz_bytes: int) -> str:
        assert sz_bytes >= 0, "sz_bytes must be non-negative"
        size, uidx = sz_bytes, 0
        units = DataUnit.units or DataUnit._get_units()
        while size >= DataUnit.KB and uidx < len(units) - 1:
            size /= float(DataUnit.KB)
            uidx += 1
        return f"{size:.2f} {units[uidx]}"





import requests, tempfile
from pathlib import Path
from . import youtube
from .files.processing import process_file
from .data import DataUnit
from .errors import NotSupportedError

def process_url(url: str) -> str:
    if youtube.is_url(url):
        return youtube.process_url(url)

    extensions = {
        "text/plain": ".txt",
        "application/pdf": ".pdf",
        "text/html": ".html"
    }

    response = requests.head(url, allow_redirects = True)
    response.raise_for_status()

    result_url = response.url
    content_type = response.headers.get("Content-Type", "")

    valid_types = list(extensions.keys())
    for ctype in valid_types:
        if content_type.startswith(ctype):
            content_type = ctype
            break

    if not content_type in valid_types:
        raise NotSupportedError(f"\nUnsupported content type detected from HEAD request: {content_type}")
    
    # send requst to download file from url (stream the data)
    response = requests.get(result_url, stream = True)
    response.raise_for_status()
    
    # create temp file and write chunks of streamed data to disk
    extension = extensions.get(content_type, ".txt")
    temp_file = tempfile.NamedTemporaryFile(suffix = extension, delete = False)
    try:
        temp_file_path = Path(temp_file.name)
        for chunk in response.iter_content(chunk_size = 8 * DataUnit.KB):
            temp_file.write(chunk)
    except Exception as ex:
        raise RuntimeError(f"Failed to write bytes to temporary file: {ex}")
    finally:
        temp_file.close()

    # process the temp file, automatically delete it after
    try:
        result = process_file(temp_file_path, delete = True)
    finally:
        if temp_file_path.exists():
            temp_file_path.unlink()
    return result



import json, re
from datetime import timedelta
from typing import List
from pathlib import Path
from youtube_transcript_api import YouTubeTranscriptApi
from . import utils, ai
from .data import DataMode
from .files import utils as FileUtils
from .serializable import Serializable
from .settings import Settings

class TranscriptEntry:

    def __init__(self, text: str, start: float, duration: float):
        self.text = text
        self.start = start
        self.duration = duration

    def __str__(self):
        start_time = str(timedelta(seconds=self.start))
        return f"{start_time} {self.text}"

    def to_dict(self):
        return {
            "text": self.text,
            "start": self.start,
            "duration": self.duration
        }

class Transcript(Serializable):

    def __init__(
        self, 
        video_id: str, 
        entries: List[TranscriptEntry], 
        vector_store_id: str = "",
        vectorize: bool = False
    ):
        self.video_id = video_id
        self.entries = entries
        self.vector_store_id = vector_store_id
        if vectorize and not vector_store_id:
            self.vectorize()

    def __str__(self):
        return "\n".join(str(entry) for entry in self.entries)

    def vectorize(self) -> ai.VectorStore:
        name = f"transcript_{self.video_id}"
        content_path = utils.get_summawise_dir() / "youtube" / f"{name}.txt"
        FileUtils.write_str(content_path, str(self))
        vector_store = ai.create_vector_store(name, [content_path])
        content_path.unlink()
        self.vector_store_id = vector_store.id
        return vector_store

    def to_json(self, pretty: bool = False) -> str:
        return json.dumps({ 
            "video_id": self.video_id,
            "entries": [entry.to_dict() for entry in self.entries],
            "vector_store_id": self.vector_store_id
        }, indent = 4 if pretty else None)

    @classmethod
    def from_json(cls, json_str: str) -> "Transcript":
        data = json.loads(json_str)
        data["entries"] = [TranscriptEntry(**entry) for entry in data.get("entries", [])]
        transcript = cls(**data)
        return transcript

def get_transcript(video_id: str, vectorize: bool = False) -> Transcript:
    transcript_data = YouTubeTranscriptApi.get_transcript(video_id)
    entries = [TranscriptEntry(**entry) for entry in transcript_data]
    return Transcript(video_id, entries, vectorize = vectorize)

def load_transcript(file_path: Path, mode: DataMode = DataMode.JSON) -> Transcript:
    # shortcut to reduce verbosity
    return Transcript.from_file(file_path, mode)

def parse_video_id(url: str) -> str:
    pattern = r"(?:v=|\/)([0-9A-Za-z_-]{11}).*"
    match = re.search(pattern, url)
    if match:
        return match.group(1)
    else:
        raise ValueError("Invalid YouTube video URL")

def is_url(url: str) -> bool:
    pattern = r"(youtu\.be|(?:www\.)?youtube\.com)\/"
    return re.search(pattern, url) is not None

def process_url(url: str) -> str:
    """
    Process a youtube URL as input.
    This function will extract a transcript, handle caching/storage, and return an OpenAI vector store ID.
    """
    # use settings class (singleton)
    settings = Settings() # type: ignore

    # file extension to cache certain objects/data (json or bin)
    ext = settings.data_mode.ext()

    video_id = parse_video_id(url)
    name = f"transcript_{video_id}"
    transcript_path = utils.fp(utils.get_summawise_dir() / "youtube" / f"{name}.{ext}")

    if not transcript_path.exists():
        # fetch transcript data from youtube
        try:
            transcript = get_transcript(video_id)
            print("Transcript retrieved and converted to text.")
        except Exception as e:
            raise Exception(f"Error retrieving transcript ({type(e).__name__}): {e}")
        
        # create vector store on openai, cache data
        try:
            vector_store_id = transcript.vectorize().id
            transcript.save_to_file(
                file_path = transcript_path,
                mode = settings.data_mode,
                compress = settings.compression
            )
            print(f"Vector store created with ID: {vector_store_id}")
        except Exception as ex:
            raise Exception(f"Error creating vector store [{type(ex)}]: {ex}")
    else:
        # restore transcript object from file and use cached vector store id
        transcript = load_transcript(transcript_path, settings.data_mode)
        vector_store_id = transcript.vector_store_id
        print(f"Restored vector store ID from cache: {vector_store_id}")

    return vector_store_id



import pickle, gzip, hashlib, chardet, xxhash
from typing import TypeVar, Type, List, Optional, NamedTuple, Union
from pathlib import Path
from enum import Enum
from .encodings import Encoding
from ..data import DataUnit
from ..utils import assert_type

T = TypeVar("T")

class HashAlg(Enum):
    SHA_256 = (hashlib, "sha256")
    SHA3_256 = (hashlib, "sha3_256")
    XXH_32 = (xxhash, "xxh32")
    XXH_64 = (xxhash, "xxh64")
    XXH_128 = (xxhash, "xxh128")
    XXH3_128 = (xxhash, "xxh3_128")
    XXH3_64 = (xxhash, "xxh3_64")

    def init(self):
        module, attr_name = self.value
        hash_init = getattr(module, attr_name)
        return hash_init() # initialze hash object

def write_str(file_path: Path, text: str, compress: bool = False) -> None:
    file_path.parent.mkdir(parents = True, exist_ok = True)
    data = text.encode("utf-8")
    write_bytes(file_path, data, compress)

def read_str(file_path: Path) -> str:
    data = read_bytes(file_path)
    text = data.decode("utf-8")
    return text

def write_bytes(file_path: Path, data: bytes, compress: bool = False) -> None:
    file_path.parent.mkdir(parents = True, exist_ok = True)
    if compress:
        if file_path.suffix != ".bin" and ".gz" not in file_path.suffixes:
            file_path = file_path.with_suffix(file_path.suffix + ".gz")
        data = gzip.compress(data)
    with open(file_path, 'wb') as file:
        file.write(data)

def read_bytes(file_path: Path) -> bytes:
    with open(file_path, 'rb') as file:
        data = file.read()
    if ".gz" in file_path.suffixes or ".bin" in file_path.suffixes:
        try:
            data = gzip.decompress(data)
        except gzip.BadGzipFile:
            pass
    return data

def save_object(file_path: Path, obj: object, compress: bool = True) -> None:
    data = pickle.dumps(obj)
    write_bytes(file_path, data, compress)

def load_object(file_path: Path, cls: Type[T]) -> T:
    obj = load_object_any(file_path)
    if not isinstance(obj, cls):
        raise TypeError(f"Expected object of type {cls.__name__}, but got {type(obj).__name__}")
    return obj

def load_object_any(file_path: Path) -> object:
    data = read_bytes(file_path)
    return pickle.loads(data)

def calculate_hash(
    _input: Union[Path, str, bytes], 
    algorithm: HashAlg = HashAlg.SHA3_256,
    intdigest: bool = False
) -> Union[str, int]:
    assert_type(_input, (bytes, str, Path))
    hash_obj = algorithm.init()

    if isinstance(_input, (bytes, str)):
        _input = _input.encode() if isinstance(_input, str) else _input
        hash_obj.update(_input)
    elif isinstance(_input, Path):
        with open(_input, "rb") as f:
            for chunk in iter(lambda: f.read(8 * DataUnit.KB), b""):
                hash_obj.update(chunk)

    return (
        hash_obj.hexdigest() if not intdigest else
        int.from_bytes(hash_obj.digest(), byteorder = "big")
    )

def list_files(directory: Path, recursive: bool = True) -> List[Path]:
    assert directory.is_dir(), f"The provided path to 'list_files' must be a directory. (Value: '{directory}')"
    files = []
    for item in directory.iterdir():
        if item.is_file():
            files.append(item)
        elif item.is_dir() and recursive:
            files.extend(list_files(item, recursive = True))
    return files

def get_encoding(file_path: Path) -> Optional[Encoding]:
    """
    Attempts to determine the encoding of a file based on the first 4 KB. If none is recognized, the function returns 'None'.
    Supported encodings: https://link.justin.ooo/chardet-encodings
    """
    with file_path.open('rb') as f:
        raw_data = f.read(4 * DataUnit.KB) # read the first 4KB of the file
        result = chardet.detect(raw_data)
        encoding_str = result.get("encoding")
        if isinstance(encoding_str, str):
            encoding = Encoding.from_string(encoding_str)
            return encoding
    return None

def has_parent_directory(path: Path, dir_name: str) -> bool:
    normalize = lambda s: f"/{s.strip('/\\').replace('\\', '/')}/"
    dir_name = normalize(dir_name)
    path_str = normalize(str(path))
    return dir_name in path_str

class FilteredFiles(NamedTuple):
    files: List[Path]
    total_count: int
    valid_count: int

def filter_files(all_files: List[Path]) -> FilteredFiles:
    encoding_whitelist = [Encoding.UTF_8, Encoding.ASCII]
    dir_blacklist = [".git", "node_modules", "site-packages"]
    extension_whitelist = {
        '.py', '.js', '.txt', '.md', '.html', '.css', '.java', '.c', '.cpp',
        '.rb', '.php', '.ts', '.json', '.xml', '.csv', '.xlsx', '.pptx', '.docx',
        '.jpg', '.jpeg', '.png', '.gif', '.webp', '.pdf', '.zip', '.tar', '.tex'
    }

    files = [
        file_path for file_path in all_files
        if file_path.suffix in extension_whitelist \
        and get_encoding(file_path) in encoding_whitelist \
        and not any(has_parent_directory(file_path, dir) for dir in dir_blacklist)
    ]

    return FilteredFiles(
        files = files,
        total_count = len(all_files),
        valid_count = len(files)
    )



from typing import Any, Union, Tuple, Optional

class NotSupportedError(Exception):
    def __init__(self, append: str = ""):
        msg = "Failed to vectorize data. Your input is not yet supported"
        msg += (": " if append else ".") + append
        super().__init__(msg)

class ValueTypeError(Exception):
    def __init__(self, value: Any, type_or_types: Union[type, Tuple[type, ...]]):
        _type: Optional[type] = None
        types: Tuple[type, ...] = ()

        if isinstance(type_or_types, tuple):
            types = type_or_types
        elif isinstance(type_or_types, type):
            _type = type_or_types
        else:
            raise TypeError("ValueTypeError 'type_or_types' must be a type, or a tuple of types.")

        msg: str = ""
        if _type is not None:
            msg = f"Expected value to be of type '{_type.__name__}', "
        elif len(types) > 0:
            types_str = " | ".join([_type.__name__ for _type in types])
            msg = f"Expected value to be of one of the types '{types_str}', "
        else:
            raise ValueError("At least one type must be specified when raising ValueTypeError.")

        msg += f"but instead received value of type '{type(value).__name__}': {value}"
        super().__init__(msg)

class MultipleAssistantsFoundError(Exception):
    def __init__(self, condition: str, value: str):
        super().__init__(f"Multiple assistants found. (Condition: '{condition}', Value: '{value}')")

class MissingSortKeyError(Exception):
    def __init__(self, append = ""):
        msg = "Missing required sort key" + (": " if len(append) else ".") + append
        super().__init__(msg)




import tempfile, sys
from datetime import datetime
from dataclasses import is_dataclass, fields
from importlib import metadata
from typing import Any, Optional, Callable, Union, Tuple, Set, Dict
from pathlib import Path
from packaging.version import Version
from .errors import ValueTypeError
from packaging.version import Version
from prompt_toolkit.document import Document
from prompt_toolkit.validation import Validator, ValidationError

package_name = lambda: __name__.split('.')[0]
utc_now = lambda: datetime.utcnow()

class Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

class NumericChoiceValidator(Validator):

    def __init__(self, valid_choices):
        self.valid_choices = valid_choices

    def validate(self, document: Document):
        try:
            value = int(document.text)
            if value not in self.valid_choices:
                raise ValidationError(
                    message="That number isn't a valid choice.",
                    cursor_position=len(document.text)
                )
        except ValueError:
            raise ValidationError(
                message="Input the number corresponding with your choice.",
                cursor_position=len(document.text)
            )

def get_summawise_dir() -> Path:
    temp_dir = Path(tempfile.gettempdir())
    return temp_dir / "summawise"

def fp(file_path: Path) -> Path:
    """
    Patch a given 'Path' object in a specific scenario:
    The correct path is the same exact location, but with a .gz suffix, denoting gzip compression.
    """
    if not file_path.exists():
        gz_path = file_path.with_suffix(file_path.suffix + ".gz")
        if gz_path.exists():
            return gz_path
    return file_path

def assert_type(value: Any, type_or_types: Union[type, Tuple[type, ...]]) -> None:
    """
    Asserts that a variable is of a given type.
    
    Parameters:
        value (Any): The variable to be checked.
        type_or_types (Union[type, Tuple[type, ...]]): The type or types that the variable should be checked against.

    Raises:
        TypeError: If 'type_or_types' is not a type or tuple of types.
        ValueTypeError: If the variable is not of the specified type or types.
    """
    if not isinstance(type_or_types, (type, tuple)):
        raise TypeError("Function 'assert_type' parameter 'type_or_types' must be a type, or a tuple of types.")
    elif not isinstance(value, type_or_types):
        raise ValueTypeError(value, type_or_types)

def asdict_exclude(obj: Any, exclude: Set[str]) -> Dict[str, Any]:
    """
    Convert a dataclass object to a dictionary, excluding specified fields.
    Parameters:
        obj (Any): The dataclass object to convert to a dictionary.
        exclude (Set[str]): A set of field names to exclude from the resulting dictionary.
    Returns:
        Dict[str, Any]: A dictionary representation of the dataclass object, excluding the specified fields.
    Raises:
        TypeError: If the input object is not a dataclass.
    """
    if not is_dataclass(obj):
        raise TypeError("Object must be a dataclass.")

    result = {}
    for f in fields(obj):
        if f.name not in exclude:
            result[f.name] = getattr(obj, f.name)
    return result

def get_version(pkg_name: str = "") -> Version:
    """
    Retrieves the version of a specified package.
    By default, it will return the running packages (summawise) version. 
    Args:
        pkg_name (str): The name of the package to retrieve the version for. If not provided, the default package name will be used.
    Returns:
        Version: An object representing the version of the specified package.
    """
    if not pkg_name: 
        pkg_name = package_name()
    version_str = metadata.version(pkg_name)
    return Version(version_str)

def delete_lines(count: int = 1):
    """
    Deletes the specified number of lines from the terminal output.
    VT100 docs: https://vt100.net/docs/vt100-ug/chapter3.html
    Parameters:
        count (int): The number of lines to delete. Default is 1.
    """
    CURSOR_UP_ONE = '\x1b[1A'
    ERASE_LINE = '\x1b[2K'
    for _ in range(count):
        sys.stdout.write(CURSOR_UP_ONE)
        sys.stdout.write(ERASE_LINE)
    sys.stdout.flush()

converter_iso: Callable[[datetime], str] = lambda v: v.isoformat()
converter_ts: Callable[[datetime], float] = lambda v: v.timestamp()
converter_ts_int: Callable[[datetime], float] = lambda v: int(v.timestamp())

def convert_datetimes(
    input_dict: Dict[str, Any], 
    converter: Callable[[datetime], Any] = converter_iso
) -> Dict[str, Any]:
    """
    Converts datetime objects in the input dictionary based on the converter func.
        - converter_iso: returns datetime.isoformat() [DEFAULT]
        - converter_ts: returns datetime.timestamp()
        - converter_ts_int: returns int(datetime.timestamp())
    Recursively applies the conversion to nested dictionaries.

    Args:
        input_dict (Dict[str, Any]): A dictionary where keys are strings and values can be of any type.

    Returns:
        Dict[str, Any]: A dict with the same keys as the input dictionary, but with datetime objects converted to ISO formatted strings.
    """
    output = {}
    for key, value in input_dict.items():
        if isinstance(value, datetime):
            output[key] = converter(value)
        elif isinstance(value, dict):
            output[key] = convert_datetimes(value, converter)
        else:
            output[key] = value
    return output

def try_parse_int(value: str, default: Optional[int] = None) -> Optional[int]:
    try:
        return int(value)
    except (ValueError, TypeError):
        return default
